package meerkat.crypto;

import com.google.protobuf.Message;

import java.io.IOException;
import java.io.InputStream;
import java.security.InvalidKeyException;
import java.security.SignatureException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.util.List;
import static meerkat.protobuf.Crypto.*;

/**
 * Created by talm on 25/10/15.
 *
 * Sign and verifyarrays of messages
 */
public interface DigitalSignature {
    final public static String CERTIFICATE_ENCODING_X509 = "X.509";

    /**
     * Load a set of certificates from an input stream.
     * This will consume the entire stream.
     * Certificates can be either DER-encoded (binary) or PEM (base64) encoded.
     * This may be called multiple times to load several different certificates.
     * It must be called before calling {@link #verify(Signature, List)}.
     * @param certStream source from which certificates are loaded
     * @throws CertificateException on parsing errors
     */
    public void loadVerificationCertificates(InputStream certStream)
            throws CertificateException;

    /**
     * Clear the loaded verification certificates.
     */
    public void clearVerificationCertificates();

    /**
     * Verify a signature on a list of messages.
     * The signature is computed on the serialized messages, separated with
     * the {@link Digest#CONCAT_MARKER} string.
     * For the signature to be valid, the verification key corresponding
     * to the signature must have been loaded using {@link #loadVerificationCertificates(InputStream)};
     * otherwise a CertificateException is thrown
     *
     * @param sig  The signature on the messages
     * @param msgs The messages themselves
     * @return
     * @throws CertificateException the certificate required for verification was not loaded.
     */
    public boolean verify(Signature sig, List<Message> msgs)
            throws CertificateException, InvalidKeyException;

    /**
     * Loads a private signing key. The certificate must be in PKCS12 format and include both
     * the public and private keys.
     * This method must be called before calling {@link #sign(List)}
     * Calling this method again will replace the key.
     *
     * @param keyStream
     */
    public void loadSigningCertificate(InputStream keyStream, char[] password)
            throws IOException, CertificateException, UnrecoverableKeyException;

    public Signature sign(List<Message> msg) throws SignatureException;

    public void clearSigningKey();

}
