package meerkat.crypto.concrete;

import java.io.IOError;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.security.*;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.*;

import com.google.protobuf.ByteString;
import meerkat.crypto.Digest;
import meerkat.protobuf.Crypto;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.protobuf.Message;

import meerkat.crypto.DigitalSignature;
import meerkat.protobuf.Crypto.Signature;
import org.bouncycastle.util.io.pem.*;
import org.bouncycastle.openssl.*;


/**
 * Sign and verify digital signatures.
 * <p>
 * This class is not thread-safe (each thread should have its own instance).
 */
public class ECDSASignature implements DigitalSignature {
    final Logger logger = LoggerFactory.getLogger(getClass());

    final public static String KEYSTORE_TYPE = "PKCS12";
    final public static String DEFAULT_SIGNATURE_ALGORITHM = "SHA256withECDSA";

    SHA256Digest digest = new SHA256Digest();

    Map<ByteString, PublicKey> loadedCertificates = new HashMap<>();

    ByteString loadedSigningKeyId = null;

    /**
     * The actual signing implementation.
     */
    java.security.Signature signer;


    /**
     * Compute a fingerprint of a public key as a SHA256 hash.
     *
     * @param pubKey
     * @return
     */
    public ByteString computePublicKeyID(PublicKey pubKey) {
        digest.reset();
        byte[] data = pubKey.getEncoded();
        digest.update(data);
        return ByteString.copyFrom(digest.digest());
    }

    public ECDSASignature(java.security.Signature signer) {
        this.signer = signer;
    }

    public ECDSASignature() {
        try {
            this.signer = java.security.Signature.getInstance(DEFAULT_SIGNATURE_ALGORITHM);
        } catch (NoSuchAlgorithmException e) {
            // Should never happen
            logger.error("Couldn't find implementation for {} signatures: {}", DEFAULT_SIGNATURE_ALGORITHM, e);
        }
    }

    @Override
    public void loadVerificationCertificates(InputStream certStream)
            throws CertificateException {
        CertificateFactory certificateFactory = CertificateFactory.getInstance(CERTIFICATE_ENCODING_X509);
        Collection<? extends Certificate> certs = certificateFactory.generateCertificates(certStream);
        for (Certificate cert : certs) {
            // Just checking
            if (!(cert instanceof X509Certificate)) {
                logger.error("Certificate must be in X509 format; got {} instead!", cert.getClass().getCanonicalName());
                continue;
            }
            PublicKey pubKey = cert.getPublicKey();
            ByteString keyId = computePublicKeyID(pubKey);
            loadedCertificates.put(keyId, pubKey);
        }
    }

    @Override
    public void clearVerificationCertificates() {
        loadedCertificates.clear();
    }


    public void updateSigner(List<Message> msgs) throws SignatureException {
        if (msgs == null)
            return;
        boolean first = true;
        for (Message msg : msgs) {
            if (!first) {
                signer.update(Digest.CONCAT_MARKER);
            }
            first = false;
            signer.update(msg.toByteString().asReadOnlyByteBuffer());
        }
    }

    public void updateSigner(InputStream in) throws IOException, SignatureException {
        ByteString inStr = ByteString.readFrom(in);
        signer.update(inStr.asReadOnlyByteBuffer());
    }

    public Signature sign() throws SignatureException {
        Signature.Builder sig = Signature.newBuilder();
        sig.setType(Crypto.SignatureType.ECDSA);
        sig.setData(ByteString.copyFrom(signer.sign()));
        sig.setSignerId(loadedSigningKeyId);
        return sig.build();
    }

    public void initVerify(Signature sig)
            throws CertificateException, InvalidKeyException  {
        PublicKey pubKey = loadedCertificates.get(sig.getSignerId());
        if (pubKey == null) {
            logger.warn("No public key loaded for ID {}!", sig.getSignerId());
            throw new CertificateException("No public key loaded for " + sig.getSignerId());
        }
        signer.initVerify(pubKey);
    }

    public boolean verify(Signature sig) {
        try {
            return signer.verify(sig.getData().toByteArray());
        } catch (SignatureException e) {
            // Should never happen!
            logger.error("Signature exception: {}", e);
            return false;
        }
    }

    @Override
    public boolean verify(Signature sig, List<Message> msgs)
            throws CertificateException {
        try {
            initVerify(sig);
            updateSigner(msgs);
            return verify(sig);
        } catch (InvalidKeyException e) {
            logger.error("Invalid key exception: {}", e);
            return false;
        } catch (SignatureException e) {
            // Should never happen!
            logger.error("Signature exception: {}", e);
            return false;
        }
    }

    @Override
    public void loadSigningCertificate(InputStream keyStream, char[] password)
            throws IOException, CertificateException, UnrecoverableKeyException {
        try {
            KeyStore keyStore = KeyStore.getInstance(KEYSTORE_TYPE);

            keyStore.load(keyStream, password);

            KeyStore.ProtectionParameter protParam =
                    new KeyStore.PasswordProtection(password);

            // Iterate through all aliases until we find the first privatekey
            Enumeration<String> aliases = keyStore.aliases();
            while (aliases.hasMoreElements()) {
                String alias = aliases.nextElement();
                try {
                    KeyStore.Entry entry = keyStore.getEntry(alias, protParam);
                    if (entry instanceof KeyStore.PrivateKeyEntry) {
                        KeyStore.PrivateKeyEntry privEntry = (KeyStore.PrivateKeyEntry) entry;
                        PrivateKey privKey = privEntry.getPrivateKey();
                        PublicKey pubKey = privEntry.getCertificate().getPublicKey();
                        loadedSigningKeyId = computePublicKeyID(pubKey);
                        signer.initSign(privKey);
                        return;
                    }
                } catch (InvalidKeyException e) {
                    logger.info("Read invalid key: {}", e);
                } catch (UnrecoverableEntryException e) {
                    logger.info("Read unrecoverable entry: {}", e);
                }

            }
        } catch (KeyStoreException e) {
            logger.error("Keystore exception: {}", e);
        } catch (NoSuchAlgorithmException e) {
            logger.error("NoSuchAlgorithmException exception: {}", e);
            throw new CertificateException(e);
        }
        throw new UnrecoverableKeyException("Didn't find valid private key entry in stream!");
    }


    @Override
    public Signature sign(List<Message> msgs) {
        try {
            updateSigner(msgs);
            return sign();
        } catch (SignatureException e) {
            logger.error("Signature exception: {}", e);
            return null;
        }
    }


    public Signature sign(InputStream in) throws IOException {
        try {
            updateSigner(in);
            return sign();
        } catch (SignatureException e) {
            logger.error("Signature exception: {}", e);
            return null;
        }
    }

    public void clearSigningKey() {
        try {
            // TODO: Check if this really clears the key from memory
            if (loadedSigningKeyId != null)
                signer.initSign(null);
            loadedSigningKeyId = null;
        } catch (InvalidKeyException e) {
            // Do nothing
        }
    }


}
